#
#  This file is part of the BayesicFitting package.
#
#  BayesicFitting is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of
#  the License, or (at your option) any later version.
#
#  BayesicFitting is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#
"""
Provides fitter functions.

Import into BayesicFitting itself all classe that are directly usable.
I.e. leave out the base classes and helper classes.

TBC: How much time does this take. Everything is translated ??

"""

from .source.AmoebaFitter import AmoebaFitter
from .source.AnnealingAmoeba import AnnealingAmoeba
from .source.ArctanModel import ArctanModel
from .source.BSplinesModel import BSplinesModel
from .source.BaseFitter import BaseFitter
from .source.BaseModel import BaseModel
from .source.BirthEngine import BirthEngine
from .source.BracketModel import BracketModel
from .source.CauchyErrorDistribution import CauchyErrorDistribution
from .source.CauchyPrior import CauchyPrior
from .source.ChebyshevPolynomialModel import ChebyshevPolynomialModel
from .source.CombiModel import CombiModel
from .source.ConstantModel import ConstantModel
from .source.ConvergenceError import ConvergenceError
from .source.CrossEngine import CrossEngine
from .source.CurveFitter import CurveFitter
from .source.DeathEngine import DeathEngine
from .source.Dynamic import Dynamic
from .source.Engine import Engine
from .source.ErrorDistribution import ErrorDistribution
from .source.EtalonDriftModel import EtalonDriftModel
from .source.EtalonModel import EtalonModel
from .source.ExpModel import ExpModel
from .source.Explorer import Explorer
from .source.ExponentialPrior import ExponentialPrior
from .source.Fitter import Fitter
from .source.FixedModel import FixedModel
#from .source.FreeShape2dModel import FreeShape2dModel
#from .source.FreeShapeModel import FreeShapeModel
from .source.GalileanEngine import GalileanEngine
from .source.GaussErrorDistribution import GaussErrorDistribution
from .source.GaussModel import GaussModel
from .source.GaussPrior import GaussPrior
from .source.GenGaussErrorDistribution import GenGaussErrorDistribution
from .source.GibbsEngine import GibbsEngine
from .source.HarmonicModel import HarmonicModel
from .source.HarmonicDynamicModel import HarmonicDynamicModel
from .source.HyperParameter import HyperParameter
from .source.ImageAssistant import ImageAssistant
from .source.IterationPlotter import IterationPlotter
from .source.IterativeFitter import IterativeFitter
from .source.JeffreysPrior import JeffreysPrior
from .source.Kernel2dModel import Kernel2dModel
from .source.KernelModel import KernelModel
from .source.LaplaceErrorDistribution import LaplaceErrorDistribution
from .source.LaplacePrior import LaplacePrior
from .source.LevenbergMarquardtFitter import LevenbergMarquardtFitter
from .source.LinearModel import LinearModel
from .source.LogFactorial import logFactorial
from .source.LorentzModel import LorentzModel
from .source.MaxLikelihoodFitter import MaxLikelihoodFitter
from .source.Model import Model
from .source.MonteCarlo import MonteCarlo
from .source.NestedSampler import NestedSampler
from .source.NoiseScale import NoiseScale
from .source.NonLinearModel import NonLinearModel
#from .source.OrderEngine import OrderEngine
from .source.PadeModel import PadeModel
from .source.PoissonErrorDistribution import PoissonErrorDistribution
from .source.PolySineAmpModel import PolySineAmpModel
from .source.PolySurfaceModel import PolySurfaceModel
from .source.PolynomialDynamicModel import PolynomialDynamicModel
from .source.PolynomialModel import PolynomialModel
from .source.PowerLawModel import PowerLawModel
from .source.PowerModel import PowerModel
from .source.Prior import Prior
from .source.ProductModel import ProductModel
from .source.QRFitter import QRFitter
from .source.RandomEngine import RandomEngine
from .source.RepeatingModel import RepeatingModel
from .source.RobustShell import RobustShell
from .source.Sample import Sample
from .source.SampleList import SampleList
from .source.ScaledErrorDistribution import ScaledErrorDistribution
## import all fitters inside ScipyFitter
from .source.ScipyFitter import *
from .source.SincModel import SincModel
from .source.SineAmpModel import SineAmpModel
from .source.SineDriftModel import SineDriftModel
from .source.SineModel import SineModel
from .source.SineSplineDriftModel import SineSplineDriftModel
from .source.SineSplineModel import SineSplineModel
from .source.SplinesModel import SplinesModel
from .source.StartEngine import StartEngine
from .source.StepEngine import StepEngine
from .source.SurfaceSplinesModel import SurfaceSplinesModel
from .source.UniformPrior import UniformPrior
from .source.VoigtModel import VoigtModel

from .source.Formatter import formatter
from .source.Formatter import formatter_init
from .source.Plotter import plotFit
from .source.Tools import printclass
from .source import Tools
#from .source import bspline
#from .source import splinelab

#from .source.bsplines.bspline import bspline
#from .source.bsplines.splinelab import splinelab

from .source.kernels.Biweight import Biweight
from .source.kernels.CosSquare import CosSquare
from .source.kernels.Cosine import Cosine
from .source.kernels.Gauss import Gauss
from .source.kernels.Huber import Huber
from .source.kernels.Kernel import Kernel
from .source.kernels.Lorentz import Lorentz
from .source.kernels.Parabola import Parabola
from .source.kernels.Sinc import Sinc
from .source.kernels.Triangle import Triangle
from .source.kernels.Tricube import Tricube
from .source.kernels.Triweight import Triweight
from .source.kernels.Uniform import Uniform




# run with : python3 -m unittest TestWeights

import unittest
import numpy as np
from numpy.testing import assert_array_almost_equal as assertAAE

import matplotlib.pyplot as plt
from BayesicFitting import *


class TestWeights( unittest.TestCase  ) :

    def plotEvidence( self ) :
        self.testEvidence( plot=True )


    def testWeights2( self, plot=None ) :
        print( "====testWeights 2====================" )
        nn = 5
        x = np.arange( nn, dtype=float )
        y = x % 2
        y[2] = 0.5
        x4 = np.arange( 4*nn, dtype=float )
        y4 = x4 % 2
        y4[range(2,20,5)] = 0.5

        pm = PolynomialModel( 0 )
        bf = CurveFitter( x, pm )
        pars = bf.fit( y )
        pcov = bf.covariance
        var = bf.makeVariance()
        print( "======= 5 data points; no weights; no fixScale ===" )
        print( "pars  ", pars, "stdv  ", np.sqrt( np.diag( pcov ) ), bf.hessian, pcov )
        print( "pars  ", pm.parameters, "stdv  ", bf.stdevs, bf.hessian, bf.covariance, var )
        print( "chisq %f  scale %f %f  sumwgt %f" % ( bf.chisq, bf.scale, bf.makeVariance(), bf.sumwgt ) )

        pm = PolynomialModel( 0 )
        bf = CurveFitter( x4, pm )
        pars = bf.fit( y4 )
        pcov = bf.covariance
        var = bf.makeVariance()
        print( "======= 20 data points; no weights; no fixScale ===" )
        print( "pars  ", pars, "stdv  ", np.sqrt( np.diag( pcov ) ), bf.hessian, pcov )
        print( "pars  ", pm.parameters, "stdv  ", bf.stdevs, bf.hessian, bf.covariance, var )
        print( "chisq %f  scale %f %f  sumwgt %f" % ( bf.chisq, bf.scale, bf.makeVariance(), bf.sumwgt ) )


        print( "======= 5 data points; weights = 4; no fixScale ===" )
        pm = PolynomialModel( 0 )
        bf = CurveFitter( x, pm )
        w = np.zeros( nn, dtype=float ) + 4
        pars = bf.fit( y, w )
        pcov = bf.covariance
        var = bf.makeVariance()
        print( "pars  ", pars, "stdv  ", np.sqrt( np.diag( pcov ) ), bf.hessian, pcov )
        print( "pars  ", pm.parameters, "stdv  ", bf.stdevs, bf.hessian, bf.covariance, var )
        print( "chisq %f  scale %f %f  sumwgt %f" % ( bf.chisq, bf.scale, bf.makeVariance(), bf.sumwgt ) )

        pm = PolynomialModel( 0 )
        bf = CurveFitter( x, pm )
        pars = bf.fit( y )
        pcov = bf.covariance
        var = bf.makeVariance()
        print( "======= 5 data points; no weights; fixScale ===" )
        print( "pars  ", pars, "stdv  ", np.sqrt( np.diag( pcov ) ), bf.hessian, pcov )
        print( "pars  ", pm.parameters, "stdv  ", bf.stdevs, bf.hessian, bf.covariance, var )
        print( "chisq %f  scale %f %f  sumwgt %f" % ( bf.chisq, bf.scale, bf.makeVariance(), bf.sumwgt ) )
        print( "noise ", bf.scale, " +- ", bf.stdevScale, "  keep ", bf.keep )


    def testWeights1( self, plot=None ) :
        print( "====testWeights 1====================" )
        nn = 5
        x = np.arange( nn, dtype=float )
        y = x % 2
        y[2] = 0.5
        x4 = np.arange( 4*nn, dtype=float )
        y4 = x4 % 2
        y4[range(2,20,5)] = 0.5

        pm = PolynomialModel( 0 )
        bf = Fitter( x, pm )
        pars = bf.fit( y )
        var = bf.makeVariance()
        print( "======= 5 data points; no weights; no noiseScale ===" )
        print( "pars  ", pm.parameters, "stdv  ", bf.stdevs, bf.hessian, bf.covariance, var )
        print( "chisq %f  scale %f %f  sumwgt %f" % ( bf.chisq, bf.scale, bf.makeVariance(), bf.sumwgt ) )

        pm = PolynomialModel( 0 )
        bf = Fitter( x4, pm )
        pars = bf.fit( y4 )
        print( "======= 20 data points; no weights; no noiseScale ===" )
        print( "pars  ", pm.parameters, "stdv  ", bf.stdevs, bf.hessian, bf.covariance )
        print( "chisq %f  scale %f %f  sumwgt %f" % ( bf.chisq, bf.scale, bf.makeVariance(), bf.sumwgt ) )


        print( "======= 5 data points; weights = 4; no noiseScale ===" )
        pm = PolynomialModel( 0 )
        bf = Fitter( x, pm )
        w = np.zeros( nn, dtype=float ) + 4
        pars = bf.fit( y, w )
        print( "pars  ", pm.parameters, "stdv  ", bf.stdevs, bf.hessian, bf.covariance )
        print( "chisq %f  scale %f %f  sumwgt %f" % ( bf.chisq, bf.scale, bf.makeVariance(), bf.sumwgt ) )

        pm = PolynomialModel( 0 )
        bf = Fitter( x, pm )
        pars = bf.fit( y )
        print( "======= 5 data points; no weights; noiseScale ===" )
        print( "pars  ", pm.parameters, "stdv  ", bf.stdevs, bf.hessian, bf.covariance )
        print( "chisq %f  scale %f %f  sumwgt %f" % ( bf.chisq, bf.scale, bf.makeVariance(), bf.sumwgt ) )
        print( "noise ", bf.scale, " +- ", bf.stdevScale, "  keep ", bf.keep )

        """
        print( "evid %f  occam %f  lhood %f" %
            ( bf.getLogZ(), bf.getLogOccam(), bf.getLogLikelihood() ) )

        ns = NoiseScale()
        ns.setLimits( [0.01, 10] )
        pm.noiseScale = ns
        print( "======= Parameters and scale" )
        print( "evid %f  occam %f  lhood %f" %
            ( bf.getLogZ(), bf.getLogOccam(), bf.getLogLikelihood() ) )

        print( "======= Parameters and fixed scale = 0.01" )
        pm.noiseScale.keepFixed( scale = 0.01 )
        print( "evid %f  occam %f  lhood %f" %
            ( bf.getLogZ(), bf.getLogOccam(), bf.getLogLikelihood() ) )

        print( "======= Parameters and fixed scale = 0.1" )
        pm.noiseScale.keepFixed( scale = 0.1 )
        print( "evid %f  occam %f  lhood %f" %
            ( bf.getLogZ(), bf.getLogOccam(), bf.getLogLikelihood() ) )

        print( "======= Parameters and fixed scale = 1.0" )
        pm.noiseScale.keepFixed( scale = 1.0 )
        print( "evid %f  occam %f  lhood %f" %
            ( bf.getLogZ(), bf.getLogOccam(), bf.getLogLikelihood() ) )


        print( "======= Parameters and fixed scale = 10.0" )
        pm.noiseScale.keepFixed( scale = 10.0 )
        print( "evid %f  occam %f  lhood %f" %
            ( bf.getLogZ(), bf.getLogOccam(), bf.getLogLikelihood() ) )
        """

if __name__ == '__main__':
    unittest.main( )




import numpy as numpy
from . import Tools
from .LinearModel import LinearModel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester

class PolySurfaceModel( LinearModel ):
    """
    General polynomial surface model of arbitrary degree.
    .. math::
        f( x,y:p ) = \sum_d \sum_k p_n * x^{d-k} * y^k )

    where the first sum is over d running from 0 to degree ( inclusive )
    and the second sum is over k running from 0 to d ( inclusive ).
    The index n is just incrementing, making all p's different.

    It is a 2-dimensional linear model.

    Examples
    --------
    poly = PolySurfaceModel( 3 )         # 3rd degree polynomial
    print poly.getNumberOfParameters( )        # 10

    Category     mathematics/Fitting

    Attributes
    ----------
    degree : int
        degree of the polynomial


    """
    def __init__( self, degree, copy=None, **kwargs ):
        """
        Polynominal surface of a certain degree.

        The number of parameters is ( degree+2 ) * ( degree+1 ) / 2

        Parameters
        ----------
        degree : int
            the degree of the polynomial.
        copy : PolySurfaceModel
            model to be copied
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.

        """
        super( PolySurfaceModel, self ).__init__( (degree+2)*(degree+1)//2, ndim=2,
                copy=copy, **kwargs )
        self.degree = degree


    def copy( self ):
        """ Copy method.  """
        return PolySurfaceModel( self.degree, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: degree
        """
        dind = {"degree": int}

        if Tools.setSingleAttributes( self, name, value, dind ) :
            pass                                            # success
        else :
            super( PolySurfaceModel, self ).__setattr__( name, value )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input values.

        The partials are the powers of x,y ( xdata ) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        nx = Tools.length( xdata[:,0] )
        part = numpy.zeros( ( nx, self.npmax ), dtype=float )
        n = 0
        for d in range( self.degree + 1 ) :
            x = numpy.ones( nx, dtype=float )
            i = 0
            while i <= d :
                part[:,n+i] = x
                x *= xdata[:,1]
                i += 1
            x = numpy.ones( nx, dtype=float )
            i = d
            while i >= 0 :
                part[:,n+i] *= x
                x *= xdata[:,0]
                i -= 1
            n += d + 1
        if parlist is None :
            return part

        return part[:,parlist]

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        bn = "PolySurface: f( x,y:p ) ="
        n = 0
        plus = ""
        for d in range( self.degree + 1 ) :
            i = d
            k = 0
            while k <= d :
                bn += "%s p_%d"%(plus,n)
                if i > 0 :
                    bn += " * x%s"%( "" if i == 1 else "^%d"%i )
                if k > 0 :
                    bn += " * y%s"%( "" if k == 1 else "^%d"%k )
                n += 1
                k += 1
                i -= 1
            plus = " +"
        return bn

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.
        Parameters  k    parameter number.

        """
        dk = 0
        while k > dk:
            dk += 1
            k -= dk
        return self.yUnit / ( ( self.xUnit[0]**k ) * ( self.xUnit[1] ** (dk-k) ) )





## run as : python3 -m unittest TestErrorDistribution

from __future__ import print_function

import numpy as numpy
from numpy.testing import assert_array_almost_equal as assertAAE
import unittest
from astropy import units
import math
import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import formatter as fmt


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2017 Do Kester

class TestErrorDistribution( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """
    SQRT2 = math.sqrt( 2 )
    LGSQ2 = math.log( SQRT2 )
    LOG2 = math.log( 2 )
    EPSILON = 1e-6

    # Define x independent variable
    x = numpy.asarray( [ -1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0], dtype=float )

    # Define y dependent variable
    data = numpy.asarray( [-11.1, -9.2, -5.1, -5.2, -1.1, 1.2, 1.1, 5.2, 5.1, 8.2, 11.2], dtype=float )

    # Define wgt weigth
    wgt = numpy.asarray( [1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=float )

    def __init__( self, name ):
        super( TestErrorDistribution, self ).__init__( name )
        numpy.set_printoptions(formatter={'float': '{: 0.3f}'.format})

    def test1( self, plot=False ) :
        print( "====test1============================" )
        nn = 10000
        x = numpy.linspace( 0, 1, nn, dtype=float )
        ym = 0.0 * x
        nf = 0.01

        model = PolynomialModel( 0 )
        model.parameters = 0.0
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        errdis = GaussErrorDistribution ( x, ym )
        print( errdis )
        for k in range( 5 ) :
            noise = numpy.random.randn( nn )
            y = ym + nf * noise
            errdis.data = y
            print( fmt( k ), fmt( nf ), fmt( errdis.getScale( model ) ) )
            nf *= 10
        allpars = [0.0, 1.0]
        if plot:
            self.ploterrdis( noise, errdis, model, allpars )

        nf = 0.01
        errdis = LaplaceErrorDistribution ( x, ym )
        print( errdis )
        for k in range( 5 ) :
            noise = numpy.random.laplace( size=nn )
            y = ym + nf * noise
            errdis.data = y
            print( fmt( k ), fmt( nf ), fmt( errdis.getScale( model ) ) )
            nf *= 10

        nf = 0.01
        errdis = CauchyErrorDistribution ( x, ym )
        cp = CauchyPrior()
        print( errdis )
        for k in range( 5 ) :
            noise = numpy.random.rand( nn )
            noise = cp.unit2Domain( noise )
            y = ym + nf * noise
            errdis.data = y
            print( fmt( k ), fmt( nf ), fmt( errdis.getScale( model ) ) )
            nf *= 10

        nf = 0.01
        errdis = GenGaussErrorDistribution ( x, ym, power=1 )
        print( errdis, "  power=1" )
        for k in range( 5 ) :
            noise = numpy.random.laplace( size=nn )
            y = ym + nf * noise
            errdis.data = y
            print( fmt( k ), fmt( nf ), fmt( errdis.getScale( model ) ) )
            nf *= 10

        nf = 0.01
        errdis = GenGaussErrorDistribution ( x, ym, power=2 )
        print( errdis, "  power=2"  )
        for k in range( 5 ) :
            noise = numpy.random.randn( nn )
            y = ym + nf * noise
            errdis.data = y
            print( fmt( k ), fmt( nf ), fmt( errdis.getScale( model ) ) )
            nf *= 10

        nf = 0.01
        power = 10
        errdis = GenGaussErrorDistribution ( x, ym, power=power )
        print( errdis, "  power=%d" % power  )
        for k in range( 5 ) :
            noise = 2 * numpy.random.rand( nn ) - 1.0
            y = ym + nf * noise
            errdis.data = y
            print( fmt( k ), fmt( nf ), fmt( errdis.getScale( model ) ) )
            nf *= 10

        allpars = [0.0, 1.0, power]
        if plot :
            self.ploterrdis( noise, errdis, model, allpars )

    def ploterrdis( self, noise, errdis, model, allpars ) :
        num_bins = 20
        plt.hist( noise, num_bins, normed=1, facecolor='g', alpha=0.5 )
        xx = numpy.linspace( -0.3, 0.3, 601, dtype=float )
        lik = numpy.zeros_like( xx )
        for k,p in enumerate( xx ) :
            allpars[0] = p
            lik[k] = errdis.logLikelihood( model, allpars )
        maxlik = numpy.max( lik )
        plt.plot( xx, numpy.exp( lik - maxlik ), 'r-' )
        plt.show()


    def testGaussErrorDistribution( self ):

        print( "=======   Test Gauss Error Distribution  ==================" )

        poly = PolynomialModel( 1 )
        param = numpy.asarray( [1,10,1], dtype=float )
        ged = GaussErrorDistribution( self.x, self.data )
        self.assertTrue( ged.acceptWeight() )

        #   data = { -11, -9, -5, -5, -1, 1, 1, 5, 5, 8, 11 }
        #   f( x ) = { -10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10 }
        #   f - d=     1   1  -1   1  -1 -1  1 -1  1  0  -1

        chisq = ged.getChisq( ged.getResiduals( poly, param[:2] ), param[2] )
        print( "chisq = %8.3f"%( chisq ) )
        logL = ged.logLikelihood( poly, param )
        altL = -0.5 * ( 11 * math.log( 2 * math.pi ) + chisq )
        print( "logL  = %8.3f  %8.3f" % ( logL, altL ) )
        assertAAE( logL, altL )

        scale = 0.1
        param[2] = scale
        logL = ged.logLikelihood( poly, param )
        altL = -11 * ( 0.5 * math.log( 2 * math.pi ) +
                    math.log( scale ) ) - 0.5 * chisq / ( scale * scale )
        print( "logL  = %8.3f  %8.3f" % ( logL, altL ) )
        assertAAE( logL, altL )

        scale = 1.0
        param[2] = scale
#        fitIndex = numpy.arange( 3 )
        fitIndex = numpy.asarray( [0,1,-1] )
        dL = ged.partialLogL( poly, param, fitIndex )
        nL = ged.numPartialLogL( poly, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        scale = 0.5
        param[2] = scale
        ged.weights = self.wgt
        dL = ged.partialLogL( poly, param, fitIndex )
        nL = ged.numPartialLogL( poly, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        scale = 1.0
        for i in range( 11 ) :
            param = numpy.asarray( [i-5,5,1], dtype=float )
            print( param, "  :  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ged.logLikelihood( poly, param ), end="" )
                param[1] += 1
            print( "" )

    def testGaussErrorDistribution2( self ):
        print( "\n=====   Test 2 Gauss Error Distribution====================" )
        gm = GaussModel( )
        gm.addModel( PolynomialModel(1) )
        param = numpy.asarray( [5, 1, -0.3, 1, 1, 1], dtype=float )
        data2 = self.data.copy( ) * 0.1
        data2[1:5] = [2.3, 5.4, 4.5, 2.0]

        print( data2 )

        ged = GaussErrorDistribution( self.x, data2 )
        Tools.printclass( ged )
        np = gm.npchain
        chisq = ged.getChisq( ged.getResiduals( gm, param[:np] ), param[np] )
        print( "chisq = %8.3f"%( chisq ) )
        logL = ged.logLikelihood( gm, param )

        print( "logL  = %8.3f  %8.3f" % ( logL, -0.5 * ( 11 * math.log( 2 * math.pi ) + chisq) ) )
        assertAAE( logL, -0.5 * ( 11 * math.log( 2 * math.pi ) + chisq) )

        scale = 0.1
        param[np] = scale
        logL = ged.logLikelihood( gm, param )
        altL = ( -11 * ( 0.5 * math.log( 2 * math.pi ) +
                    math.log( scale ) ) - 0.5 * chisq / ( scale * scale ) )
        print( "logL  = %8.3f  %8.3f" % ( logL, altL ) )
        assertAAE( logL, altL )

        scale = 1.0
        param[np] = scale
        fitIndex = numpy.append( numpy.arange( np ), [-1] )
        dL = ged.partialLogL( gm, param, fitIndex )
        nL = ged.numPartialLogL( gm, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        fitIndex = numpy.asarray( [0,2,3,4] )
        dL = ged.partialLogL( gm, param, fitIndex )
        nL = ged.numPartialLogL( gm, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        scale = 0.5
        param[np] = scale
        fitIndex = numpy.asarray( [0,1,2,3,4,-1] )
        ged.weights = self.wgt
        dL = ged.partialLogL( gm, param, fitIndex )
        nL = ged.numPartialLogL( gm,  param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        scale = 0.5
        param[np] = scale
        fitIndex = numpy.asarray( [0,1,2,4,-1] )
        ged.weights = self.wgt
        ged.hyperpar[0] = NoiseScale()
        Tools.printclass( ged )
        ged.keepFixed( fixed={0:2.5} )
        Tools.printclass( ged )
        Tools.printclass( ged.hyperpar[0] )

        dL = ged.partialLogL( gm, param, fitIndex )
        nL = ged.numPartialLogL( gm, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

    def testGenGaussErrorDistribution1( self ):

        print( "=======   Test Gen Gauss Error Distribution 1 ==================" )

        poly = PolynomialModel( 1 )
        param = numpy.asarray( [1,10], dtype=float )
        ged = LaplaceErrorDistribution( self.x, self.data )
        ggd = GenGaussErrorDistribution( self.x, self.data )
        self.assertTrue( ggd.acceptWeight() )

        #   data = { -11, -9, -5, -5, -1, 1, 1, 5, 5, 8, 11 }
        #   f( x ) = { -10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10 }
        #   f - d=     1   1  -1   1  -1 -1  1 -1  1  0  -1

        power = 1.0
        ggscl = 1.0
        lpscl = 1.0
        hypar = [ggscl, power]
        ggsigma = ggd.toSigma( hypar )
        lpsigma = ged.toSigma( lpscl )
        sigma = ggsigma / lpsigma
        print( ggsigma, lpsigma, sigma )

        ff = 1.0 / sigma
        ggpar = numpy.append( param, hypar )
        lppar = numpy.append( param, [lpscl] )
        res = ggd.getResiduals( poly, param )
        chisq = ged.getSumRes( res, lpscl )
        chipp = ggd.getChisq( res, ggscl, power )
        print( "chipp = %8.3f   chisq = %8.3f"%( chipp, chisq ) )
#        ggfi = numpy.arange( 4 )
#        lpfi = numpy.arange( 3 )
        ggfi = numpy.asarray( [0,1,-2,-1] )
        lpfi = numpy.asarray( [0,1,-1] )
        logL = ged.logLikelihood( poly, lppar )
        lggL = ggd.logLikelihood( poly, ggpar )

        print( "lggL  = %8.3f   logL  = %8.3f" % ( lggL, logL ) )
        assertAAE( logL, lggL )

        dL = ged.partialLogL( poly, lppar, lpfi )
        nL = ged.numPartialLogL( poly, lppar, lpfi )
        dG = ggd.partialLogL( poly, ggpar, ggfi)
        nG = ggd.numPartialLogL( poly, ggpar, ggfi )

        print( "partial = ", dL )
        print( "partial = ", dG )
        print( "numpart = ", nL )
        print( "numpart = ", nG )
        assertAAE( dG[:3], dL, 5 )
        assertAAE( dG, nG, 5 )
        assertAAE( dL, nL, 5 )

        ggscl = 0.1
        hypar = [ggscl, power]
        sigma = ggscl / ff
        ggpar = numpy.append( param, hypar )
        lppar = numpy.append( param, [sigma] )
        logL = ged.logLikelihood( poly, lppar )
        lggL = ggd.logLikelihood( poly, ggpar )
        print( "lggL  = %8.3f   logL  = %8.3f" % ( lggL, logL ) )
        assertAAE( logL, lggL )

        ggscl = 0.5
        hypar = [ggscl, power]
        sigma = ggscl / ff
        print( "Super : ", hypar, "  sigma : ", sigma )
        ggpar = numpy.append( param, hypar )
        lppar = numpy.append( param, [sigma] )

        ged.weights = self.wgt
        ggd.weights = self.wgt
        dL = ged.partialLogL( poly, lppar, lpfi )
        nL = ged.numPartialLogL( poly, lppar, lpfi )
        dG = ggd.partialLogL( poly, ggpar, ggfi)
        nG = ggd.numPartialLogL( poly, ggpar, ggfi )

        print( "partial = ", dL )
        print( "partial = ", dG )
        print( "numpart = ", nL )
        print( "numpart = ", nG )
        assertAAE( dG[:3], dL, 5 )
        assertAAE( dG, nG, 5 )
        assertAAE( dL, nL, 5 )

        for i in range( 11 ) :
            param = numpy.asarray( [i-5, 5, ggscl, power], dtype=float )
            print( param, "  :  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ggd.logLikelihood( poly, param ), end="" )
                param[1] += 1
            print( "" )

    def testGenGaussErrorDistribution2( self ):

        print( "=======   Test Gen Gauss Error Distribution 2  ==================" )

        poly = PolynomialModel( 1 )
        param = numpy.asarray( [1,10], dtype=float )
        ged = GaussErrorDistribution( self.x, self.data )
        ggd = GenGaussErrorDistribution( self.x, self.data )
        self.assertTrue( ggd.acceptWeight() )

        #   data = { -11, -9, -5, -5, -1, 1, 1, 5, 5, 8, 11 }
        #   f( x ) = { -10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10 }
        #   f - d=     1   1  -1   1  -1 -1  1 -1  1  0  -1

        power = 2.0
        scale = 1.0
        hypar = [scale, power]
        sigma = ggd.toSigma( hypar )
        print( hypar, sigma )
        ff = 1 / sigma
        ggpar = numpy.append( param, hypar )
        gapar = numpy.append( param, [sigma] )
        res = ggd.getResiduals( poly, param )
        chisq = ged.getChisq( res, scale )
        chipp = ggd.getChisq( res, scale, power )
        print( "chipp = %8.3f   chisq = %8.3f"%( chipp, chisq ) )
        logL = ged.logLikelihood( poly, gapar )
        lggL = ggd.logLikelihood( poly, ggpar )

        print( 'gapar  ', gapar )
        print( 'ggpar  ', ggpar )

        print( "lggL  = %8.3f   logL  = %8.3f" % ( lggL, logL ) )
        assertAAE( logL, lggL )

        ggfi = numpy.array( [0,1,-2,-1] )
        lpfi = numpy.array( [0,1,-1] )

        dL = ged.partialLogL( poly, gapar, lpfi )
        nL = ged.numPartialLogL( poly, gapar, lpfi )
        dG = ggd.partialLogL( poly, ggpar, ggfi)
        nG = ggd.numPartialLogL( poly, ggpar, ggfi )

        print( "partial = ", dL )
        print( "numpart = ", nL )
        print( "partial = ", dG )
        print( "numpart = ", nG )
        print( nG[-2] * ff )

        assertAAE( dG[:2], dL[:2], 5 )
        assertAAE( dG[2]*ff, dL[2], 5 )
        assertAAE( dG, nG, 5 )
        assertAAE( dL, nL, 5 )

        scale = 0.1
        hypar = [scale,power]
        sigma = ggd.toSigma( hypar )
        ggpar = numpy.append( param, hypar )
        gapar = numpy.append( param, [sigma] )
        logL = ged.logLikelihood( poly, gapar )
        lggL = ggd.logLikelihood( poly, ggpar )

        print( "lggL  = %8.3f   logL  = %8.3f" % ( lggL, logL ) )
        assertAAE( logL, lggL )

        dL = ged.partialLogL( poly, gapar, lpfi )
        nL = ged.numPartialLogL( poly, gapar, lpfi )
        dG = ggd.partialLogL( poly, ggpar, ggfi)
        nG = ggd.numPartialLogL( poly, ggpar, ggfi )

        print( "partial = ", dL )
        print( "partial = ", dG )
        print( "numpart = ", nL )
        print( "numpart = ", nG )
        assertAAE( dG[:2], dL[:2], 5 )
        assertAAE( dG[2]*ff, dL[2], 5 )
        assertAAE( dG, nG, 5 )
        assertAAE( dL, nL, 4 )

        scale = 0.5
        hypar = [scale, power]
        sigma = ggd.toSigma( hypar )
        print( "Super : ", hypar, "  sigma : ", sigma )
        ggpar = numpy.append( param, hypar )
        gapar = numpy.append( param, [sigma] )
        ged.weights = self.wgt
        ggd.weights = self.wgt

        logL = ged.logLikelihood( poly, gapar )
        lggL = ggd.logLikelihood( poly, ggpar )

        print( "lggL  = %8.3f   logL  = %8.3f" % ( lggL, logL ) )
        assertAAE( logL, lggL )

        dL = ged.partialLogL( poly, gapar, lpfi )
        nL = ged.numPartialLogL( poly, gapar, lpfi )
        dG = ggd.partialLogL( poly, ggpar, ggfi)
        nG = ggd.numPartialLogL( poly, ggpar, ggfi )

        print( "partial = ", dL )
        print( "partial = ", dG )
        print( "numpart = ", nL )
        print( "numpart = ", nG )
        assertAAE( dG[:2], dL[:2], 5 )
        assertAAE( dG[2]*ff, dL[2], 5 )
        assertAAE( dG, nG, 5 )
        assertAAE( dL, nL, 5 )

        for i in range( 11 ) :
            param = numpy.asarray( [i-5,5,scale,power], dtype=float )
            print( param, "  :  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ggd.logLikelihood( poly, param ), end="" )
                param[1] += 1
            print( "" )

    def testPoissonErrorDistribution( self ):
        print( "====== Test Poisson Error Distribution ======================" )
        poly = PolynomialModel( 1 )
        param = numpy.asarray( [12, 10], dtype=float )
        data = numpy.asarray( self.data + 12, dtype=int )
        print( "Data : ", data )
        ped = PoissonErrorDistribution( self.x, data )
        self.assertFalse( ped.acceptWeight() )

        logL = ped.logLikelihood( poly, param )
        print( "logL  = %8.3f"%( logL ) )
        scale = 0.1
        logL = ped.logLikelihood( poly, param )
        mok = poly.result( self.x, param )
        altL = numpy.sum( data * numpy.log( mok ) - mok - logFactorial( data ) )

        print( "logL  = %8.3f  %8.3f" % ( logL, altL ) )
        assertAAE( logL, altL )

        logL = ped.logLikelihood( poly, [-5, 0] )
        print( "logL  = %8.3f"%( logL ) )
        self.assertTrue( math.isinf( logL ) )

        scale = 1.0
        fitIndex = [0,1]
        dL = ped.partialLogL( poly, param, fitIndex )
        nL = ped.numPartialLogL( poly, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        scale = 0.5
        dL = ped.partialLogL( poly, param, fitIndex )
        nL = ped.numPartialLogL( poly, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        scale = 1.0
        for i in range( 10 ) :
            param = numpy.asarray( [8+i,4], dtype=float )
            print( param, "  :  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ped.logLikelihood( poly, param ), end="" )
                param[1] += 1
            print( "" )


    def testCauchyErrorDistribution( self ):
        print( "\n   Test Cauchy Error Distribution\n" )
        poly = PolynomialModel( 1 )
        ced = CauchyErrorDistribution( self.x, self.data )
        self.assertFalse( ced.acceptWeight() )

        param = numpy.asarray( [1,10,1], dtype=float )
        logL = ced.logLikelihood( poly, param )
        print( "logL  =  %8.3f" % ( logL ) )
        scale = 0.1
        s2 = scale * scale
        param[2] = scale
        logL = ced.logLikelihood( poly, param )
        altL = math.log( scale / math.pi ) * len( self.data )
        res = ced.getResiduals( poly, param[:2] )
        altL -= numpy.sum( numpy.log( s2 + res * res ) )
        print( "logL  =  %8.3f  alt %8.3f" % (logL, altL ) )
        assertAAE( logL, altL )

        scale = 1.0
        param[2] = scale
        fi = [0,1,-1]
        dL = ced.partialLogL( poly, param, fi )
        nL = ced.numPartialLogL( poly, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

        scale = 0.5
        param[2] = scale
        dL = ced.partialLogL( poly, param, fi )
        nL = ced.numPartialLogL( poly, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )


        for i in range( 11 ):
            param = numpy.asarray( [i-5,5,1], dtype=float )
            print( param, ":  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ced.logLikelihood( poly, param ), end="" )
                param[1] += 1
            print( "" )


    def testLaplaceErrorDistribution( self ):
        print( "\n   Test Laplace Error Distribution\n" )
        poly = PolynomialModel( 1 )
        ced = LaplaceErrorDistribution( self.x, self.data )
        self.assertTrue( ced.acceptWeight() )

        param = numpy.asarray( [1,10,1], dtype=float )
        logL = ced.logLikelihood( poly, param )
        print( "logL  =  %8.3f" % ( logL ) )
        scale = 0.1
        param[2] = scale
        s2 = scale * scale
        logL = ced.logLikelihood( poly, param )

        altL = - len( self.data ) * math.log( 2.0 * scale )
        res = ced.getResiduals( poly, param[:2] )
        altL -= numpy.sum( numpy.abs( res ) ) / scale
        print( "logL  =  %8.3f  alt %8.3f" % (logL, altL ) )
        assertAAE( logL, altL )

        scale = 1.0
        param[2] = scale
        fi = [0,1,-1]
        dL = ced.partialLogL( poly, param, fi )
        nL = ced.numPartialLogL( poly, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

        scale = 0.5
        param[2] = scale
        ced.weights = self.wgt
        dL = ced.partialLogL( poly, param, fi )
        nL = ced.numPartialLogL( poly, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

        logL = ced.logLikelihood( poly, param )
        cced = ced.copy()
        logLc = cced.logLikelihood( poly, param )
        print( cced )
        print( "logL  =  %8.3f  copy %8.3f" % (logL, logLc ) )
        dLc = cced.partialLogL( poly, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        assertAAE( logL, logLc, 6 )
        assertAAE( dL, dLc, 6 )

        for i in range( 11 ):
            param = numpy.asarray( [i-5,5,1], dtype=float )
            print( param, ":  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ced.logLikelihood( poly, param ), end="" )
                param[1] += 1
            print( "" )


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( ErrorDistributionTest.__class__ )





import math

from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester


class LaplacePrior( Prior ):
    """
    Laplace prior distribution.
    .. math::
        Pr( x ) = exp( - |x| / scale )

    By default: scale = 1.
    It can also have a limited domain. By default the domain is [-Inf,+Inf].
    In computational practice the domain is limited to about [-36,36] scale units

    Equivalent to a double-sided exponential prior

    domain2unit: u = 0.5 * exp( d / scale )             if d < 0
                     1.0 - 0.5 * exp( -d / scale )      otherwise
    unit2domain: d = log( 2 * u ) * scale               if u < 0.5
                    -log( 2 * ( 1 - u ) ) * scale       otherwise

    Attributes
    ----------
    scale : float
        scale of the exponential
    lowLimit : float
        low limit (inactive for now)
    highLimit : float
        high limit ( inactive for now)

    """

    #  *********CONSTRUCTOR***************************************************
    def __init__( self, scale=1.0, prior=None ):
        """
        Constructor.

        Parameters
        ----------
        scale : float
            of the exponential
        prior : LaplacePrior
            prior to copy (with new scale if applicable)

        """
        super( LaplacePrior, self ).__init__( prior=prior )
        self.scale = scale

    def copy( self ):
        return LaplacePrior( prior=self, scale=self.scale )

    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for a Laplace distribution.

        domain2unit: u = 0.5 * exp( d / scale ) if d < 0 else
                         1.0 - 0.5 * exp( -d / scale )

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        return ( 0.5 * math.exp(  dval / self.scale ) if ( dval < 0 ) else
           1.0 - 0.5 * math.exp( -dval / self.scale ) )

    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for a Laplace distribution.

        unit2domain: d = log( 2 * u ) * scale if u < 0.5 else
                        -log( 2 * ( 1 - u ) ) * scale;

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        if uval == 0 : return -math.inf
        elif uval == 1 : return math.inf
        scl = self.scale
        if uval > 0.5:
            uval = 1 - uval
            scl *= -1
        return math.log( 2 * uval ) * scl

    def result( self, x ):
        """
        Return a the result of the distribution function at x.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        return 0.5 * math.exp( -abs( x ) / self.scale ) / self.scale

    def logResult( self, x ):
        """
        Return a the log of the result of the distribution function to p.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        if self.isOutOfLimits( x ) : return -math.inf

        return math.log( 0.5 / self.scale ) - abs( x ) / self.scale

    def partialLog( self, p ):
        """
        Return partial derivative of log( Prior ) wrt parameter.

        Parameters
        ----------
        p : float
            the value

        """
        if p == 0 : return 0
        return -1 / self.scale if p > 0 else 1 / self.scale

    def isBound( self ):
        """ Return true if the integral over the prior is bound.  """
        return True

    def __str__( self ):
        """ Return a string representation of the prior.  """
        return str( "Laplace prior with scale = %.2f"%( self.scale ) )






import numpy as numpy
from . import Tools
from .Model import Model
from .NonLinearModel import NonLinearModel
from astropy import units

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018        Do Kester

class ParallelPipe( Model ):
    """
    Two or more models are processing in parallel, yielding 2-dim output which can/must
    be processed by a more dimensional model.

    ParallelPipe is not a fittable model by itself. It can be part of a model chain.

    The number of parameters is the sum of the parameters of the models.

    Examples
    --------
    >>> nxk = 17
    >>> xknots = numpy.arange(  nxk , dtype=float ) * 10      # make knots from 0 to 160
    >>> sm = SplinesModel( xknots )
    >>> pp = ParallelPipe( [None, sm] )
    >>> print csm.getNumberOfParameters( )      # ( nxk + order - 1 )
    19
    >>> xy = numpy.arange( 320, dtype=float ).reshape( 160, 2 )
    >>> par = numpy.ones( 19, dtype=float )
    >>> out = pp.result( xy, par )
    >>> print( out.shape )
    ( 160, 2 )


    Category     mathematics/Fitting

    Attributes
    ----------

    """

    def __init__( self, models, copy=None, fixed=None, **kwargs ):
        """
        Direct product of 2 (or more) models. It has dimensionality equal to
        the number of constituent models.

        The models are given as input the consecutive colums in xdata.

        The number of parameters is the sum of the parameters of the
        constituent models

        Parameters
        ----------
        models : list of Model or None
            the constituent models
        copy : ParallelPipe
            model to be copied
        fixed : dict
            If not None raise AttributeError.

        Raises
        ------
        ValueError
            When one of the models is 2 (ore more) dimensional
        AttributeErrr : When fixed is not None

        """
        if fixed is not None :
            raise AttributeError( "ParallelPipe cannot have fixed parameters" )

        np = 0
        for m in models :
            if m.ndim > 1 :
                raise ValueError( "Only 1-dim models are allowed in ParallelPipe" )
            np += m.npchain

        super( ParallelPipe, self ).__init__( np, ndim=len( models ), copy=copy, **kwargs )
        self.models = models

    def copy( self ):
        """ Copy method.  """
        mdls = [m.copy() for m in self.models]
        return ParallelPipe( mdls, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: models

        """
        dlst = {'models': (Model,None) }
        if Tools.setListOfAttributes( self, name, value, dlst ) :
            pass
        else :
            super( ParallelPipe, self ).__setattr__( name, value )

    def baseResult( self, xdata, params ):
        """
        Returns the partials at the input value.

        The partials are the powers of x (input) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters to the model.

        """
        n = 0
        res = xdata.copy()
        for k,m in enumerate( self.models ) :
            if m is None :
                continue
            res[:,k] = m.result( xdata[:,k], params[n:n+m.npchain] )
            n += m.npchain
        return res


    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        The partials are the powers of x (input) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters to the model.
        parlist : array_like
            not used in this model

        """
        ndata = Tools.length( xdata[:,0] )
        partial = numpy.ndarray( ( ndata, 0 ), dtype=float )
        n = 0
        for m in self.models :
            if m is None :
                continue
            nm = n + m.npchain
            i = n
            while parlist[i] < nm :
                i == 1
            pl = parlist[n:i] - n
            p = m.partial( xdata[:,k], params[n:nm], parlist=pl )
            partial = numpy.append( partial, p, axis=1 )
            n = nm
        return partial

    def baseDerivative( self, xdata, params ):
        """
        Returns the partials at the input value.

        The partials are the powers of x (input) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters to the model.

        """
        ndata = Tools.length( xdata[:,0] )
        dfdx = numpy.ndarray( ( ndata, 0 ), dtype=float )
        n = 0
        for m in self.models :
            if m is None :
                continue
            nm = n + m.npchain
            p = m.derivative( xdata[:,k], params[n:nm] )
            dfdx = numpy.append( dfdx, p, axis=1 )
            n = nm
        return dfdx

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        strm = ""
        ch = "xyzuvw"
        k = 0
        for m in self.models :
            strm += m.shortName() + "(%s) * " % ch[k]
            k += 1
        strm = strm[:-3]

        return str( "%dd-Product: f(%s:p) = %s"%(self.ndim, ch[:k], strm) )

    def baseParameterName( self, k ):
        """
        Return the name of a parameter as "param<dim>_<seq>.
        Parameters
        ----------
        k : int
            the kth parameter.

        """
        strpar = "param"
        m = 0
        for mdl in self.models :
            nx = mdl.npchain
            if k < nx :
                return strpar + "%d_%d"%(m,k)
            k -= nx
            m += 1
        return strpar

    def baseParameterUnit( self, k ):
        """
        Return the unit of a parameter.
        Parameters
        ----------
        k : int
            the kth parameter.

        """
        u = units.Unit( 1.0 )
        n = 0
        for mdl in self.models :
            nx = mdl.npbase
            if k < nx :
                return mdl.getParameterUnit( k ) / u
            n += 1
            k -= nx
            u = self.yUnit
        return u





# run with : python3 -m unittest TestBaseFitter

import unittest
import math
import numpy as np
from numpy.testing import assert_array_almost_equal as assertAAE

import matplotlib.pyplot as plt

from BayesicFitting import *

class TestBaseFitter( unittest.TestCase  ) :

    def testExceptions( self ) :
        print( "====testExceptions======================" )
        x1 = np.arange( 10, dtype=float )
        x2 = np.ndarray( [10,2] )
        pm = PolynomialModel( 1 )
        gm = GaussModel()
        pm += gm

        self.assertRaises( ValueError, BaseFitter, x1, gm )
        self.assertRaises( ValueError, BaseFitter, x2, pm )

        bf = BaseFitter( x1, pm )
        y = np.arange( 10, dtype=float )
        self.assertRaises( NotImplementedError, bf.fit, y )

        self.assertRaises( AttributeError, bf.__getattr__, 'sumwgt' )
        self.assertRaises( AttributeError, bf.__getattr__, 'chisq' )
        self.assertRaises( AttributeError, bf.__getattr__, 'logOccam' )
        self.assertRaises( AttributeError, bf.__getattr__, 'logLikelihood' )
        self.assertRaises( AttributeError, bf.__getattr__, 'xyz' )
        bf.chisq = -1.0
        print( 'chisq', bf.chisq )

        y[2] = -np.inf
        w = np.ones( 10, dtype=float )
        self.assertRaises( ValueError, bf.checkNan, y, weights=w )
        w[8] = np.nan
        y[2] = 0
        self.assertRaises( ValueError, bf.checkNan, y, weights=w )

        x1[3] = np.nan
        self.assertRaises( ValueError, BaseFitter, x1, pm )

    def testInit( self ) :
        print( "====testInit======================" )
        x1 = np.arange( 10, dtype=float )
        pm = PolynomialModel( 1 )
        bf = BaseFitter( x1, pm )
        print( bf )
        print( bf.model )
        print( bf.xdata )
        print( bf.nxdata, bf.ndim )

    def testVector( self ) :
        print( "====testVector======================" )
        x = np.arange( 10, dtype=float )
        y = np.random.rand( 10 )
        w = np.ones( 10, dtype=float )
        pm = PolynomialModel( 1 )
        bf = BaseFitter( x, pm )
        v1 = bf.getVector( y )
        print( v1 )
        v2 = bf.getVector( y, index=[0] )
        print( v2 )
        v3 = bf.getVector( y, index=[1] )
        print( v3 )

        assertAAE( v1, np.append( v2, v3 ) )

    def testHessian( self ) :
        print( "====testHessian======================" )
        x = np.arange( 10, dtype=float )
        pm = PolynomialModel( 1 )
        bf = BaseFitter( x, pm )
        y = np.ones( 10, dtype=float )
        w = np.ones( 10, dtype=float )
        p = np.asarray( [1.0,1.0] )

        h0 = bf.hessian
        print( h0 )
        h1 = bf.getHessian( )
        print( h1 )
        h2 = bf.getHessian( params=p )
        print( h2 )
        h3 = bf.getHessian( weights=w, params=p )
        print( h3 )
        assertAAE( h0, h1 )
        assertAAE( h1, h2 )
        assertAAE( h1, h3 )
        index = 0
        h4 = bf.getHessian( weights=w, params=p, index=index )
        print( h4 )
        index = 1
        h4 = bf.getHessian( weights=w, params=p, index=index )
        print( h4 )
        index = [0,1]
        w /= 2
        h4 = bf.getHessian( weights=w, params=p, index=index )
        print( h4 )
        print( bf.sumwgt )

        ih = bf.getInverseHessian()
        print( ih )

        dia = np.inner( h1, ih )
        print( dia )
        assertAAE( dia, np.eye( 2 ) )

    def testChiSquared( self ) :
        print( "====testChiSquared======================" )
        x = np.arange( 10, dtype=float )
        pm = PolynomialModel( 1 )
        bf = BaseFitter( x, pm )
        y = np.ones( 10, dtype=float )
        w = np.ones( 10, dtype=float )

        print( bf.model.parameters )
        c1 = bf.chiSquared( y )
        print( c1, bf.sumwgt )
        self.assertTrue( c1 == bf.chisq )
        self.assertTrue( bf.sumwgt == 10 )

        w /= 2
        c2 = bf.chiSquared( y, weights=w )
        print( c2, bf.sumwgt )

        cov = bf.covariance
        print( cov )

        s1 = bf.stdevs
        print( s1 )

    def testEvidence( self ) :
        print( "====testEvidence======================" )
        x = np.arange( 11, dtype=float ) - 5
        pm = PolynomialModel( 1 )
        bf = Fitter( x, pm )
        np.random.seed( 12345 )
        y = 0.2 + 0.3 * x + np.random.randn( 11 ) * 0.1
        w = np.ones( 11, dtype=float )

        par = bf.fit( y, w )
        print( "chisq %f  scale %f  sumwgt %f" % ( bf.chisq, bf.scale, bf.sumwgt ) )

        lolim = [-10.0,-10.0]
        hilim = [+10.0,+10.0]


        print( "= 0 ====== Parameters only" )

        evi0 = bf.getLogZ( limits=[lolim,hilim] )
        occ0 = bf.logOccam
        lhd0 = bf.logLikelihood
        print( "evid %f  occam %f  lhood %f" % ( evi0, occ0, lhd0 ) )
        self.assertTrue( bf.getEvidence( limits=[-10.0,10.0] ) == evi0 / math.log( 10.0 ) )

        nslim = [0.01,10.0]
        print( "= 1 ====== Parameters and scale" )
        evi1 = bf.getLogZ( limits=[-10.0,10.0], noiseLimits=nslim )
        occ1 = bf.logOccam
        lhd1 = bf.logLikelihood
        print( "evid %f  occam %f  lhood %f" % ( evi1, occ1, lhd1 ) )
        self.assertTrue( bf.fixedScale is None  )
        self.assertRaises( AttributeError, bf.__getattr__, "minimumScale" )

        print( "= 2 ====== Parameters and fixed scale = 0.01" )
        bf.fixedScale = 0.01
        evi2 = bf.getLogZ( limits=[lolim,hilim] )
        occ2 = bf.logOccam
        lhd2 = bf.logLikelihood
        print( "evid %f  occam %f  lhood %f" % ( evi2, occ2, lhd2 ) )
        self.assertFalse( bf.fixedScale is None )
        self.assertRaises( AttributeError, bf.__getattr__, "minimumScale" )

        print( "= 3 ====== Parameters and fixed scale = 0.1" )
        bf.fixedScale = 0.1
        evi3 = bf.getLogZ( limits=[lolim,hilim] )
        occ3 = bf.logOccam
        lhd3 = bf.logLikelihood
        print( "evid %f  occam %f  lhood %f" % ( evi3, occ3, lhd3 ) )

        print( "= 4 ====== Parameters and fixed scale = 1.0" )
        bf.fixedScale = 1.0
        evi4 = bf.getLogZ( limits=[lolim,hilim] )
        occ4 = bf.logOccam
        lhd4 = bf.logLikelihood
        print( "evid %f  occam %f  lhood %f" % ( evi4, occ4, lhd4 ) )

        print( "= 5 ====== Parameters and fixed scale = 10.0" )
        bf.fixedScale = 10.0
        evi5 = bf.getLogZ( limits=[lolim,hilim] )
        occ5 = bf.logOccam
        lhd5 = bf.logLikelihood
        print( "evid %f  occam %f  lhood %f" % ( evi5, occ5, lhd5 ) )
        self.assertFalse( bf.fixedScale is None )

        # no noiseLimit is the same as fixedScale=1.0
        self.assertTrue( evi0 == evi4 )
        self.assertTrue( occ0 == occ4 )
        self.assertTrue( lhd0 == lhd4 )

        # evidence with the `right' fixed scale is the best; better than auto scale
        self.assertTrue( evi3 > evi1 )      # auto scaling
        self.assertTrue( evi3 > evi2 )      # scale = 0.01
        self.assertTrue( evi3 > evi4 )      # scale = 1.0
        self.assertTrue( evi3 > evi5 )      # scale = 10.0

        print( "======= limits inside the model " )
        pm.setLimits( lolim, hilim )
        print( "limits par[0]  ", pm.priors[0].lowLimit, pm.priors[0].highLimit )
        print( "limits par[1]  ", pm.priors[1].lowLimit, pm.priors[1].highLimit )
        print( "ranges         ", pm.priors[0]._range, pm.priors[1]._range )
        self.assertTrue( pm.priors[0]._range == 20.0 )
        self.assertTrue( pm.priors[1]._range == 20.0 )

        bf.fixedScale = None
        evi6 = bf.getLogZ( )
        occ6 = bf.logOccam
        lhd6 = bf.logLikelihood
        print( "evid %f  occam %f  lhood %f" % ( evi0, occ0, lhd0 ) )
        self.assertTrue( bf.getEvidence( ) == evi0 / math.log( 10.0 ) )

        # no noiseLimit is the same as fixedScale=1.0
        self.assertTrue( evi0 == evi6 )
        self.assertTrue( occ0 == occ6 )
        self.assertTrue( lhd0 == lhd6 )


#    def testNullModel( self ) :
#        print( "====testNullModel======================  TBD" )


if __name__ == '__main__':
    unittest.main( )




import numpy as numpy
from .LinearModel import LinearModel
from . import Tools

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (JAVA Code)
#  *    2016      Do Kester


class PolynomialModel( LinearModel ):
    """
    General polynomial model of arbitrary degree.
    .. math::
        f( x:p ) = \sum p_k * x^k

    where the sum is over k running from 0 to degree ( inclusive ).

    It is a linear model.
    Examples
    --------
    >>> poly = PolynomialModel( 3 )            # 3rd degree polynomial
    >>> print( poly.getNumberOfParameters() )
    4

    Author : Do Kester

    """
    def __init__( self, degree, copy=None, **kwargs ):
        """
        Polynomial of a certain degree.

        The number of parameters is ( degree + 1 )

        Parameters
        ----------
        degree : int
            the degree of the polynomial.
        copy : PolynomialModel
            model to copy

        """
        super(PolynomialModel,self ).__init__( degree + 1, copy=copy,
                        **kwargs )

        object.__setattr__( self, "degree", degree )

    def copy( self ):
        """ Copy method.  """
        return PolynomialModel( self.degree, copy=self )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        The partials are the powers of x ( xdata ) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters for the model (ignored for LinearModels).
        parlist : array_like
            list of indices of active parameters

        """
        nxdata = Tools.length( xdata )
        x = numpy.ones( nxdata )

        partial = numpy.zeros( ( nxdata, self.npmax ), dtype=float )
        k = 0
        for i in range( self.npmax ) :
            partial[:,k] = x
            k += 1
            x *= xdata

        if parlist is None :
            return partial

        return partial[:,parlist]

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each input (=x).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters for the model.

        """
        if self.npmax <= 1 :
            return numpy.zeros_like( xdata )

        np = self.npmax
        p = params * numpy.arange( np )
        np -= 2
        return PolynomialModel( np ).result( xdata, p[1:] )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        bn = "Polynomial: f( x:p ) = p_0"
        if self.npmax > 1 :
            bn += " + p_1 * x"
        for k in range( 2, self.npmax ) :
            bn += " + p_%d * x^%d"%(k,k)
        return bn

    def baseParameterName( self, k ):
        """
        Return the name of the indicated parameter.
        Parameters
        ----------
        k : int
            parameter number.

        """
        return "polycoeff_%d" % k

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.

        """
        return self.yUnit / ( self.xUnit ** k )





import numpy as numpy
from astropy import units
import math
from .import Tools

from .HarmonicModel import HarmonicModel
from .Dynamic import Dynamic
from .Prior import Prior
from .ExponentialPrior import ExponentialPrior
from .UniformPrior import UniformPrior

__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2018        Do Kester


class HarmonicDynamicModel( HarmonicModel, Dynamic ):
    """
    Harmonic oscillator Model of adaptable order.

    f( x:p ) = &sum;_j ( p_k cos( 2 &pi; j x ) + p_k+1 sin( 2 &pi; j x ) )

    j = 1, N; k = 0, 2N, 2

    The parameters are initialized at 1.0. It is a linear model.

    Author       Do Kester

    Examples
    --------
    >>> harm = HarmonicDynamicModel( 3 )     # period = 1
    >>> print harm.getNumberOfParameters( )        # 6
    >>> harm = HarmonicModel( 4, 2.7 )        # period = 2.7

    Category     mathematics/Fitting

    """
    deltaNpar = 2

    def __init__( self, order, minOrder=1, maxOrder=None, period=1.0, fixed=None,
                  growPrior=None, copy=None, **kwargs ):
        """
        Harmonic of a adaptable order.

        The model starts as a HarmonicModel of order = 1
        Growth of the model is governed by a exponential prior ( scale=1 )

        Parameters
        ----------
        order : int
            order to start with. It should be minOrder <= order <= maxOrder
        minOrder : int
            minimum degree of polynomial (def=1)
        maxOrder : None or int
            maximum degree of polynomial (def=None)
        period : float
            period of the oscilation
        growPrior : None or Prior
            governing the birth and death.
            ExponentialPrior (scale=2) if  maxOrder is None else UniformPrior
        copy : HarmonicDynamicModel
            model to copy

        Raises
        ------
        AttributeError when fixed parameters are requested
        ValueError when order is outside [min..max] range

        """
        if fixed is not None :
            raise AttributeError( "DynamicModel cannot have fixed parameters" )
        if order < minOrder or ( maxOrder is not None and order > maxOrder ) :
            raise ValueError( "order outside range of [min..max] range" )

        super( HarmonicDynamicModel, self ).__init__( order, period=period,
            copy=copy, **kwargs )

        if copy is None :
            self.minComp = minOrder
            self.maxComp = maxOrder
            if growPrior is None :
                if maxOrder is None :
                    self.growPrior = ExponentialPrior( scale=2 )
                else :
                    lim = [minOrder, maxOrder+1]        # limits on components
                    self.growPrior = UniformPrior( limits=lim )
            else :
                self.growPrior = growPrior
        else :
            self.minComp = copy.minComp
            self.maxComp = copy.maxComp
            self.growPrior = copy.growPrior.copy()

    def copy( self ):
        """ Copy method.  """
        return HarmonicDynamicModel( self.order, period=self.period, copy=self )

    def changeNComp( self, dn ) :
        self.order += dn

    def __setattr__( self, name, value ) :
        dind = {"minComp": int, "maxComp": int, "growPrior": Prior}
        lnon = {"maxComp": int}
        if ( Tools.setNoneAttributes( self, name, value, lnon ) or
             Tools.setSingleAttributes( self, name, value, dind ) ) :
            pass
        else :
            super( HarmonicDynamicModel, self ).__setattr__( name, value )


    def baseName( self ):
        """ Return a string representation of the model.  """
        return "Dynamic" + super( HarmonicDynamicModel, self ).baseName( )





# run with : python3 -m unittest TestFormatter

import unittest
import numpy as numpy
import math

from BayesicFitting import formatter as fmt
from BayesicFitting import formatter_init as fmtinit

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2006-2015 Do Kester (JAVA CODE)

class TestFormatter( unittest.TestCase ) :
    """
    Test harness for Fitter class.

    @author Do Kester

    """
    def test1( self ):
        print( "===== formatter test1 ===========================" )

        arr = numpy.asarray( [k for k in range(120)], dtype=float  )
        arr = arr.reshape( (3,40) )

        print( fmt( arr, max=None ) )
        print( "arr", fmt( arr[1], indent=4, format=" %7.2f", max=20 ) )
        print( fmt( arr.reshape( (3,4,10) ), max=2 ) )
        alist = [1,2,3,4,5]
        print( fmt( alist ) )
        self.assertTrue( isinstance( alist, list ) )
        print( fmt( 3 ), fmt( 3.4 ) )

    def test2( self ):
        print( "===== formatter test2 ===========================" )

        arr = numpy.asarray( [k for k in range(36)], dtype=float  )
        arr = arr.reshape( (3,12) )

        fmtinit( max=None )
        print( fmt( arr, max=None ) )
        fmtinit( linelength=60 )
        print( fmt( arr, max=None ) )
        fmtinit( linelength=80, format={"float64" : " %7.2f"} )
        print( fmt( arr, max=None ) )

        print( "arr", fmt( arr, indent=4, format=" %7.2f" ) )
        print( fmt( arr, max=2 ) )
        alist = [1,2,3,4,5]
        print( fmt( alist ) )
        self.assertTrue( isinstance( alist, list ) )
        print( fmt( 3 ), fmt( 3.4 ) )

    def suite( cls ):
        return ConfiguredTestCase.suite( TestFormatter.__class__ )

if __name__ == '__main__':
    unittest.main( )





import math
from scipy import special
from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017        Do Kester


class GaussPrior( Prior ):
    """
    Gauss prior distribution.
    .. math::
        Pr( x ) = exp( - ( x / scale )^2 )

    By default: scale = 1.
    It can also have a limited domain. By default the domain is [-Inf,+Inf].
    In computational practice the domain is limited to about [-6, 6] scale units.

    Equivalent to a double-sided exponential prior

    domain2unit: u = 0.5 * ( erf( d / scale ) + 1 )
    unit2domain: d = erfinv( 2 * u - 1 ) * scale

    Attributes
    ----------
    scale : float
        scale of the Gaussian distribution
    lowLimit : float
        low limit (inactive for now)
    highLimit : float
        high limit ( inactive for now)

    """

    SPI = 1.0 / math.sqrt( math.pi )
    LSPI = math.log( SPI )

    #  *********CONSTRUCTOR***************************************************
    def __init__( self, scale=1.0, prior=None ):
        """
        Constructor.

        Parameters
        ----------
        scale : float
            of the exponential
        prior : GaussPrior
            prior to copy (with new scale if applicable)

        """
        super( GaussPrior, self ).__init__( prior=prior )
        self.scale = scale

    def copy( self ):
        return GaussPrior( prior=self, scale=self.scale )

    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for a Gauss distribution.

        domain2unit: u = 0.5 * ( erf( d / scale ) + 1 )

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        return 0.5 * ( special.erf(  dval / self.scale ) + 1 )

    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for a Gauss distribution.

        unit2domain: d = erfinv( 2 * u - 1 ) * scale

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        return special.erfinv( 2 * uval - 1 ) * self.scale

#    def partialDomain2Unit( self, dval ):
    def result( self, x ):
        """
        Return a the result of the distribution function at x.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        xs = x / self.scale
        return self.SPI * math.exp( - xs * xs )

    def logResult( self, x ):
        """
        Return a the log of the result of the distribution function to p.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        xs = x / self.scale
        return self.LSPI - ( xs * xs )

    def partialLog( self, x ):
        """
        Return partial derivative of log( Prior ) wrt x.

        Parameters
        ----------
        x : float
            the value

        """
        return -2 * x / ( self.scale * self.scale )

    def isBound( self ):
        """ Return true if the integral over the prior is bound.  """
        return True

    def __str__( self ):
        """ Return a string representation of the prior.  """
        return str( "Gauss prior with scale = %.2f"%( self.scale ) )






# run with : python3 -m unittest TestStartEngine

import unittest
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt

from TestEngine import TestEngine
from BayesicFitting import *



__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestStartEngine( TestEngine ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """
    def __init__( self, name ):
        super( TestStartEngine, self ).__init__( name )


    def plotStartEngine( self ):
        self.testStartEngine( plot=True )

    def testStartEngine( self, plot=False ):
        print( "\n   Start Engine Test 1\n" )
        m, xdata, data = self.initEngine()

        errdis = GaussErrorDistribution( xdata, data )
        errdis.setLimits( [0.1, 10.0] )

        sl = SampleList( m, 100, errdis )

        engine = StartEngine( sl, errdis )
        self.startenginetest( engine )

        print( "    make copy of engine" )
        copeng = engine.copy()
        self.startenginetest( copeng )

        if plot :
            parevo = sl.getParameterEvolution()
            sclevo = sl.getScaleEvolution()
            print( parevo.shape, sclevo.shape )
            plt.subplot( 2, 2, 1 )
            plt.plot( parevo[:,0], parevo[:,1], 'k.' )
            plt.subplot( 2, 2, 2 )
            plt.plot( parevo[:,0], parevo[:,2], 'k.' )
            plt.subplot( 2, 2, 3 )
            plt.plot( parevo[:,1], parevo[:,2], 'k.' )
            plt.subplot( 2, 2, 4 )
            plt.plot( parevo[:,0], sclevo, 'k.' )
            plt.show()

    def startenginetest( self, engine ) :
        print( engine )
        minv = [ math.inf] * len( engine.walkers[0].allpars )
        maxv = [-math.inf] * len( engine.walkers[0].allpars )
        for samp in engine.walkers :
            engine.execute( samp, -math.inf )
            minv = numpy.fmin( minv, samp.allpars )
            maxv = numpy.fmax( maxv, samp.allpars )
            if samp.id <= 1 or samp.id >= 98 :
                print( samp )
                print( "  ", samp.allpars, "  ", samp.fitIndex )
        print( "minv  ", minv )
        print( "maxv  ", maxv )


    def suite( cls ):
        return unittest.TestCase.suite( TestStartEngine.__class__ )


if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
from astropy import units
import math
from . import Tools
from .LinearModel import LinearModel
from .SplinesModel import SplinesModel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017        Do Kester

class SineSplineModel( LinearModel ):
    """
    Sine of fixed frequency with splineslike amplitudes/phases.

    .. math::
        f( x:p ) = s_0 \cos( 2 \pi \omega x ) + s_1 \sin( 2 \pi \omega x )

    Where :math:`s_0` and :math:`s_1` are splines with defined knots and order.
    It is a linear model with 2 * ( len(knots) + order - 1 ) papameters.

    Examples
    --------
    >>> knots = [3.0*k for k in range( 11 )]
    >>> sine = SineSplineModel( 150, knots )        # fixed frequency of 150 Hz
    >>> print( sine.npbase )                        # number of parameters
    26
    """

    def __init__( self, frequency, knots, order=3, copy=None, fixed=None, **kwargs ):
        """
        Sine model of a fixed frequency with a splineslike changing amplitude/phase.

        Number of parameters is 2 * ( len(knots) + order - 1 ).

        Parameters
        ----------
        frequency : float
            the frequency
        copy : SineSplineModel
            model to be copied
        fixed : dict
            If not None raise AttributeError.

        Raises
        ------
        AttributeError
            when fixed is not None

        """
        if fixed is not None :
            raise AttributeError( "SineSplineModel cannot have fixed parameters" )

        np = 2 * ( len( knots ) + order - 1 )
        super( SineSplineModel, self ).__init__( np, copy=copy, **kwargs )

        self.frequency = frequency
        self.knots = knots
        self.order = order
        if copy is not None :
            self.cm = copy.cm.copy()
            self.sm = copy.sm.copy()
        else :
            self.cm = SplinesModel( knots, order=order )
            self.sm = SplinesModel( knots, order=order )

    def copy( self ):
        """ Copy method.  """
        return SineSplineModel( self.frequency, self.knots, order=self.order, copy=self )

    def __setattr__( self, name, value ) :
        dind = {"frequency": float, "order": float, "cm": SplinesModel, "sm": SplinesModel}
        dlst = {"knots": float}
        if not ( Tools.setListOfAttributes( self, name, value, dlst ) or
                 Tools.setSingleAttributes( self, name, value, dind ) ) :
            super( SineSplineModel, self ).__setattr__( name, value )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters of the model (ignored in LinearModels)
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        nxdata = Tools.length( xdata )
        part = numpy.zeros( ( nxdata, self.npbase ), dtype=float )
        nh = self.npbase // 2
        cx = numpy.cos( 2 * math.pi * self.frequency * xdata )
        sx = numpy.sin( 2 * math.pi * self.frequency * xdata )

        part[:,:nh] = ( cx * self.cm.basePartial( xdata, params ).transpose() ).transpose()
        part[:,nh:] = ( sx * self.sm.basePartial( xdata, params ).transpose() ).transpose()
        return part

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative of f to x (df/dx) at the input value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters of the model

        """
        tpf = 2 * math.pi * self.frequency
        tx = tpf * xdata
        cx = numpy.cos( tx )
        sx = numpy.sin( tx )
        amps = self.getAmplitudes( xdata, params )
        cadx = self.cm.baseDerivative( xdata, params )
        sadx = self.sm.baseDerivative( xdata, params )
        return tpf * ( cadx * cx - amps[0] * sx + sadx * sx + amps[1] * cx )

    def getAmplitudes( self, xdata, params ) :
        """
        Return the amplitudes if cosine and sine, resp.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters of the model

        """
        nh = self.npbase // 2
        return ( self.cm.result( xdata, params[:nh] ),
                 self.sm.result( xdata, params[nh:] ) )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return ( "SineSpline: f( x:p ) = spline_0 * cos( 2 pi * x * f ) + " +
                 "spline_1 * sin( 2 pi * x * f ); f = %f"%self.frequency )

    def baseParameterUnit( self, k ):
        """
        Return the name of a parameter.
        Parameters
        ----------
        k : int
            the kth parameter.

        """
        k = k % ( self.npbase / 2 )
        if k > self.order :
            k = self.order
        return self.yUnit / ( self.xUnit ** k )






import numpy as numpy
import math

from .Kernel import Kernel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  *    This file is part of the BayesicFitting package.
#  *
#  *    BayesicFitting is free software: you can redistribute it and/or modify
#  *    it under the terms of the GNU Lesser General Public License as
#  *    published by the Free Software Foundation, either version 3 of
#  *    the License, or ( at your option ) any later version.
#  *
#  *    BayesicFitting is distributed in the hope that it will be useful,
#  *    but WITHOUT ANY WARRANTY; without even the implied warranty of
#  *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  *    GNU Lesser General Public License for more details.
#  *
#  *    The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 - 2017 Do Kester

class Huber( Kernel ):
    """
    Huber is an improper Kernel function which is constant between [-1,1]
    and 1 / |x| elsewhere.

    It is improper because the integral equals +inf.

    It plays a role in robust fitting, see RobustShell, for medianizing the fit.

    """
    def __init__( self ) :
        super( Huber, self ).__init__( integral=math.inf, fwhm=4.0 )

    def result( self, x ):
        ax = numpy.abs( x )
        return numpy.where( ax < 1, 1.0, 1.0 / ax )

    def resultsq( self, xsq ):
        ax = numpy.sqrt( xsq )
        return numpy.where( ax < 1, 1.0, 1.0 / ax )

    def partial( self, x ):
        ax = numpy.abs( x )
        return numpy.where( ax < 1, 0.0, -1.0 / ( x * ax ) )

    def isBound( self ):
        return False

    def name( self ):
        return str( "Huber: 1 if |x| < 1 else 1/|x|" )





# run with : python3 -m unittest TestEngine3

import unittest
import numpy as numpy
from astropy import units
import math

import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestEngine3( unittest.TestCase ):
    """
    Test harness for Engines

    Author       Do Kester

    """
    def __init__( self, name ):
        super( TestEngine3, self ).__init__( name )


    def plotRandomEngine( self ):
        self.testRandomEngine(plot=True )

    def testRandomEngine( self, plot=False ):
        print( "\n   Random Engine Test\n" )
        self.stdenginetest( RandomEngine, nsamp=10, plot=plot )

    def plotGibbsEngine( self ):
        self.testGibbsEngine( plot=True )

    def testGibbsEngine( self, plot=False ):
        print( "\n   Gibbs Engine Test\n" )
        self.stdenginetest( GibbsEngine, nsamp=10, plot=plot )

    def plotGibbsEngine2( self ):
        self.testGibbsEngine2( plot=True )

    def testGibbsEngine2( self, plot=False ):
        print( "\n   Gibbs Engine Test 2\n" )
        self.stdenginetest2( GibbsEngine, nsamp=10, lowL=-200, plot=plot )
        self.stdenginetest2( GibbsEngine, nsamp=10, lowL=-170, plot=plot )
        self.stdenginetest2( GibbsEngine, nsamp=10, lowL=-165, plot=plot )
        self.stdenginetest2( GibbsEngine, nsamp=10, lowL=-163, plot=plot )

    def plotStepEngine( self ):
        self.testStepEngine( plot=True )

    def testStepEngine( self, plot=False ):
        print( "\n   Step Engine Test 2\n" )
        self.stdenginetest2( StepEngine, nsamp=10, lowL=-200, plot=plot )
        self.stdenginetest2( StepEngine, nsamp=10, lowL=-170, plot=plot )
        self.stdenginetest2( StepEngine, nsamp=10, lowL=-165, plot=plot )
        self.stdenginetest2( StepEngine, nsamp=10, lowL=-163, plot=plot )

    def plotGalileanEngine( self ):
        self.testGalileanEngine( plot=True )

    def testGalileanEngine( self, plot=False ):
        print( "\n   Galilean Engine Test\n" )
        self.stdenginetest( GalileanEngine, nsamp=10, plot=plot )

    def plotGalileanEngine2( self ):
        self.testGalileanEngine2( plot=True )

    def testGalileanEngine2( self, plot=False ):
        print( "\n   Galilean Engine Test 2\n" )
        self.stdenginetest2( GalileanEngine, nsamp=10, lowL=-200, plot=plot )
        self.stdenginetest2( GalileanEngine, nsamp=10, lowL=-170, plot=plot )
        self.stdenginetest2( GalileanEngine, nsamp=10, lowL=-165, plot=plot )
        self.stdenginetest2( GalileanEngine, nsamp=10, lowL=-163, plot=plot )

    def initEngine( self, order=1, np=101 ):
        m = PolynomialModel( order )

        up1 = UniformPrior( limits=[-1.0,1.0])
        up2 = UniformPrior( limits=[-1.0,1.0] )

        m.priors = [up1, up2]

        xdata = numpy.linspace( -1.0, 1.0, np, dtype=float )
        data = -0.4 * xdata + 0.5

        numpy.random.seed( 345345 )
        numpy.set_printoptions( precision=3, suppress=True )
        return ( m, xdata, data )

    def stdenginetest( self, myengine, nsamp=4, plot=False ) :
        m, xdata, data = self.initEngine()
        sigma = 2.0
        errdis = GaussErrorDistribution( xdata, data )
        errdis.keepFixed( {0: sigma} )
        sl = SampleList( m, nsamp, errdis )
        map = numpy.ndarray( (21,21), dtype=float )
        mmx = -math.inf
        for k0 in range( 21 ) :
            for k1 in range( 21 ) :
                pl = [k0/10-1, k1/10-1, sigma]
                map[k1,k0] = errdis.logLikelihood( m, pl )
                if map[k1,k0] > mmx :
                    mmx = map[k1,k0]
                    mk1 = k1/10 - 1
                    mk0 = k0/10 - 1

#        print( map )
#        print( mk0, mk1, mmx )

        ay = numpy.linspace( -1, 1, 21 )
        ax = numpy.linspace( -1, 1, 21 )
        v = [-200, -170, -165, -163]

        lowL = -170

        engine = StartEngine( sl, errdis )
        for samp in engine.walkers :
            while True :
                engine.execute( samp, -math.inf )
                if samp.logL > lowL : break

        engine = myengine( sl, errdis )
        engine.plotter = Plotter()

        pevo = sl.getParameterEvolution()
        for k in range( 5 ) :
            engine.plotter.iter = 1
            plt.figure( str( myengine ) )
            plt.contour( ax, ay, map, v )
            plt.plot( [mk0], [mk1], 'k*' )

            klo = engine.rng.randint( 0, nsamp )

            p0 = sl[klo].allpars[:2]
            plt.plot( p0[0], p0[1], 'k.' )
            engine.calculateUnitRange()

            engine.execute( sl[klo], lowL )
            p1 = sl[klo].allpars[:2]

            print( klo, p1, sl[klo].logL )
            plt.show()


    def stdenginetest2( self, myengine, nsamp=4, lowL=-170, plot=False ) :
        m, xdata, data = self.initEngine()
        sigma = 2.0
        errdis = GaussErrorDistribution( xdata, data )
        errdis.keepFixed( {0: sigma} )
        sl = SampleList( m, nsamp, errdis )

        engine = StartEngine( sl, errdis )
        for samp in engine.walkers :
            while True :
                engine.execute( samp, -math.inf )
                if samp.logL > lowL : break

        pevo = sl.getParameterEvolution()
#        print( fmt( pevo ) )
        pmin = numpy.min( pevo, 0 )
        pmax = numpy.max( pevo, 0 )
        pran = ( pmax - pmin ) / 2
#        print( fmt( pmin ), fmt( pmax ), fmt( pran ) )
        pmin -= pran
        pmax += pran

        map = numpy.ndarray( (21,21), dtype=float )
        mmx = -math.inf
#        print( fmt(pmin), fmt(pmax) )
#        print( fmt(m.priors[0].lowLimit), fmt(m.priors[0].highLimit) )
#        print( fmt(m.priors[1].lowLimit), fmt(m.priors[1].highLimit) )

        ax = numpy.linspace( max( pmin[0], m.priors[0].lowLimit ),
                min( pmax[0], m.priors[0].highLimit ), 21 )
        ay = numpy.linspace( max( pmin[1], m.priors[1].lowLimit ),
                min( pmax[1], m.priors[1].highLimit ), 21 )

        for k0,p0 in enumerate( ax ) :
            for k1,p1 in enumerate( ay ) :
                pl = [p0, p1, sigma]
                map[k1,k0] = errdis.logLikelihood( m, pl )
                if map[k1,k0] > mmx :
                    mmx = map[k1,k0]
                    mk1 = p1
                    mk0 = p0

#        print( map )
        print( mk0, mk1, mmx )



        plt.figure( str( myengine ) )
        v = [-200, -170, -165, -163]
        plt.contour( ax, ay, map, v )
#        plt.plot( [mk0], [mk1], 'k*' )
        plt.xlabel( "Param[0]" )
        plt.ylabel( "Param[1]" )

        engine = myengine( sl, errdis )
        klo = engine.rng.randint( 0, nsamp )

        p0 = sl[klo].allpars[:2]
        engine.calculateUnitRange()

        for k in range( 100 ) :
            sl[klo].allpars[:2] = p0
            engine.execute( sl[klo], lowL )
            p1 = sl[klo].allpars[:2]
            plt.plot( p1[0], p1[1], 'r.' )

        plt.plot( p0[0], p0[1], 'k*' )
        plt.show()

    def suite( cls ):
        return unittest.TestCase.suite( TestEngine2.__class__ )


class Plotter( object ) :

    def __init__( self, iter=1 ) :
        self.iter = iter
        self.col = ['k-', 'r-', 'g-', 'b-', 'm->']

    def start( self ):
        """ start the plot. """
        pass

    def move( self, param, ptry, col=None ):
        """
        Move parameters at position param to ptry using color col.
        """
        plt.plot( [param[0], ptry[0]], [param[1], ptry[1]], self.col[col] )
        plt.text( ptry[0], ptry[1], "%d"%self.iter )
        self.iter += 1

    def stop( self ):
        """ Stop (show) the plot. """
        pass


if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
from astropy import units
import re
import warnings
from . import Tools

from .BaseModel import BaseModel

__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *           2017 Do Kester

class FixedModel( BaseModel ):
    """
    A FixedModel is a BaseModel where some parameters are permanently fixed.

    A parameter can be fixed either with a constant float value or more
    dynamically, with another Model. The parameters of this latter model
    also appear as parameters of the FixedModel.

    The methods `result` (f(x:p)) and `partial` (df/dp) are calculated
    in here.
    Unfortunately the methods `derivative` (df/dx) is model dependent.
    It is reset to `numDerivative`.

    Examples
    --------
    >>> m1 = PolynomialModel( 1 )
    >>> m1 += SineModel()
    >>> print( m1.npchain )         # number of params: 2 + 3
    5
    >>> fixed = { 0: 1.0, 1: m1 }
    >>> em = EtalonModel( fixed=fixed )
    >>> print( em.npbase, em.npmax, em.npchain )          # ( 4 - 2 ) + 5
    7 9 7
    >>> print( em )

    Attributes
    ----------
    npmax : int
        maximum number of parameters of the simple (not-fixed) model
    fixed : dictionary of {int:float|Model}
        int     index of parameter to fix permanently. Default None.
        float   value for the fixed parameter.
        Model   model to replace the parameter
        Attribute fixed can only be set in the constructor.
    parlist : array_like or None
        list of active (not-fixed) indices. None is all.


    Author :         Do Kester

    """

    #  *************************************************************************
    def __init__( self, nparams=0, ndim=1, copy=None, fixed=None,
                  names=None, **kwargs ) :
        """
        FixedModel Constructor.

        Parameters
        ----------
        nparams : int
            Number of parameters in the model (default: 0)
        ndim : int
            Number of dimensions of the input (default: 1)
        copy : BaseModel
            to be copied
        fixed : dictionary of {int:float|Model}
            int     index of parameter to fix permanently. Default None.
            float   value for the fixed parameter.
            Model   model to replace the parameter
            Attribute fixed can only be set in the constructor.
        names : list of string
            names for the parameters

        """

        mlist = []
        npar = nparams
        if fixed is not None :
            for k in fixed.keys() :
                if isinstance( fixed[k], BaseModel ) :
                    mlist += [k]
                    npar += fixed[k].npchain

        super( FixedModel, self ).__init__( nparams=npar, ndim=ndim,
                copy=copy, **kwargs )


        if copy is None :
            self.npmax = nparams
            if fixed is not None :
                self.npbase -= len( fixed )
                self._npb = npar
            self.mlist = mlist
            self._setFixed( fixed )
            if names is None :
                self.parNames = ["parameter_%d"%k for k in range( self.npbase )]
            else :
                self.parNames = self.selectNames( names )
        else :
            self.npmax = copy.npmax
            self.mlist  = copy.mlist
            if copy.fixed is not None :
                fixed = {}
                self._npb = copy._npb
                for k in copy.fixed.keys() :
                    if k in self.mlist :
                        fixed[k] = copy.fixed[k].copy()
                    else :
                        fixed[k] = copy.fixed[k]
            else :
                fixed = None
            self._setFixed( fixed )
            self.parNames = copy.parNames[:]

    def copy( self ) :
        """ Return a copy. """
        return FixedModel( copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes.

        """
        key1 = {"npmax":int, "_npb":int, "fixed":dict }
        keys = {"mlist":int}
        if ( Tools.setListOfAttributes( self, name, value, keys ) or
             Tools.setSingleAttributes( self, name, value, key1 ) ) :
            pass
        elif name == "fixed" :
            raise AttributeError( "Attribute fixed can only be set in the constructor" )
        else :
            super( FixedModel, self ).__setattr__( name, value )

    def _setFixed( self, fixed ) :
        object.__setattr__( self, "fixed", fixed )
        if fixed is None :
            object.__setattr__( self, "parlist", None )
            return

        if not isinstance( fixed, dict ) :
            raise AttributeError( "Attribute fixed needs to be a dictionary" )

        parlist = numpy.arange( self._npb )
        parlist = numpy.setxor1d( parlist, list( fixed.keys() ) )
        object.__setattr__( self, "parlist", parlist )

    def select( self, params ) :
        """
        Select the relevant parameters and store them.

        Parameters
        ----------
        params : array of float
            parameters of the head model
        """

        if self.fixed is None :
            return params

        # only fixed floats; np models
        if len( self.mlist ) == 0 :
            return params[self.parlist]

        pl = numpy.intersect1d( self.parlist, numpy.arange( self.npmax ) )
        pars = params[pl]
        for m in self.mlist :
            pars = numpy.append( pars, self.fixed[m].parameters )
        return pars

    def selectNames( self, names ) :
        """
        Select the relevant parameter names and store them.

        Parameters
        ----------
        names : list of string
            parameter names of the head model
        """
        if self.fixed is None :
            return names
        if len( self.mlist ) == 0 :
            return [names[k] for k in self.parlist]

        pl = numpy.intersect1d( self.parlist, numpy.arange( self.npmax ) )
        nms = [names[k] for k in pl]

        for m in self.mlist :
            mdl = self.fixed[m]
            while mdl is not None:
                nms += mdl.parNames
                mdl = mdl._next
        return nms

    #  *****RESULT**************************************************************
    def result( self, xdata, param ):
        """
        Returns the result calculated at the xdatas.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        expparam = self.expand( xdata, param )
        return super( FixedModel, self ).result( xdata, expparam )

    def expand( self, xdata, param ) :
        """
        Returns a complete list of parameters, where the fixed parameters
        have been replaced by either a constant value or by the results of
        the fixed function.
        """
        if self.fixed is None :
            return param

        pars = numpy.zeros( self._npb, dtype=float )
        pars[self.parlist] = param
        if len( self.mlist ) == 0 :
            fl =  list( self.fixed.keys() )
            pars[fl] = list( self.fixed.values() )
            return pars

        par = pars.tolist()
        ns = self.npmax - len( self.fixed )
        for k in list( self.fixed.keys() ) :
            if k in self.mlist :
                nf = ns + self.fixed[k].npchain
                par[k] = self.fixed[k].result( xdata, param[ns:nf] )
                ns = nf
            else :
                par[k] = self.fixed[k]

        ## par is now a heterogeneous list of floats and [floats]
        ## It can only be cast into an array of object

        return numpy.asarray( par, dtype=object )

    #  *****PARTIAL*************************************************************
    def partial( self, xdata, param ):
        """
        Returns the partial derivatives calculated at the inputs.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        expparam = self.expand( xdata, param )
        self.checkParameter( expparam )
        if len( self.mlist ) == 0 :
            return super(FixedModel, self).partial( xdata, expparam, parlist=self.parlist )

        np = len( param )
        ns = len( expparam )
        nf = self.npmax - len( self.fixed )
        plist = numpy.append( self.parlist[:nf], self.mlist )
        partial = numpy.zeros( ( Tools.length( xdata ), np ), dtype=float )

        bp = self.basePartial( xdata, expparam[:self.npmax], parlist=plist )
        partial[:,:nf] = bp[:,:nf]
        ns = nf
        nb = self.npmax
        for k in self.mlist :
            ne = self.fixed[k].npchain
            pa = expparam[nb:nb+ne]
            pf = self.fixed[k].partial( xdata, pa )
            ppp = bp[:,nf] * pf.transpose()
            partial[:,ns:ns+ne] = ppp.transpose()
            ns += ne
            nf += 1
            nb += ne
        return partial

    def numPartial( self, xdata, params, parlist=None ) :
        """
        Returns numerical partial derivatives of the model to params.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters; default is self.parameters
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        nxdata = Tools.length( xdata )
        partial = numpy.zeros( ( nxdata, self.npbase ), dtype=float  )
        dpg = Tools.makeNext( self.deltaP, 0 )

        par = params.copy()
        if parlist is None :
            parlist = numpy.arange( self.npbase )

        i = 0

        for k in parlist :
            dp = next( dpg )
            par[k] += 0.5 * dp
            r1 = FixedModel.result( self, xdata, par )
            par[k] -= dp
            r2 = FixedModel.result( self, xdata, par )
            partial[:,i] = ( r1 - r2 ) / dp
#            print( k, i, params[k], dp, par[k], r1, r2, partial[:,i] )
            par[k] = params[k]
            i += 1
        try :
            dpg.close()
        except :
            pass
        return partial

    def basePartial( self, xdata, param, parlist=None ) :
        """
        Replacement for models that dont define a partial.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        param : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        warnings.warn( self.shortName() + " has no partials defined; " +
                    "using numeric partials." )

        if parlist is None :
            return FixedModel.numPartial( self, xdata, param, parlist=parlist )

        par = param[parlist]
        pl = numpy.arange( len( par ), dtype=int )
        return FixedModel.numPartial( self, xdata, par, parlist=pl )

    def derivative( self, xdata, param ) :
        """
        Returns the derivative of the model to xdata.

        It is a numeric derivative as the analytic derivative is not present
        in the model.

        If `fixed` contains a Model, the derivative cannot be constructed
        from the constituent models. Use `numDerivative` instaed.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the derivative
        param : array_like
            values for the parameters. (default: self.parameters)

        """
        if len( self.mlist ) == 0 :
            params = self.expand( xdata, param )
            return self.baseDerivative( xdata, params )
        else :
            return FixedModel.numDerivative( self, xdata, param )

    def numDerivative( self, xdata, param ) :
        """
        Returns the numeric derivative of the model to input

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the derivative
        param : array_like
            values for the parameters. (default: self.parameters)

        Raises
        ------
        ValueError when the number of xdata dimensions > 1.

        """
        if self.ndim > 1 :
            raise ValueError( "Not implemented for dimensions > 1" )

        dx = self.deltaP[0]
        xd = xdata + dx
        r1 = FixedModel.result( self, xd, param )
        xd = xdata - dx
        r2 = FixedModel.result( self, xd, param )

        return ( r1 - r2 ) / ( 2 * dx )


    #  *****TOSTRING***********************************************************
    def _toString( self, npars=0 ) :
        basename = self.baseName()

        if self.fixed is not None :
            for k in self.fixed.keys() :
                par = "p_%d "%k
                if k in self.mlist :
                    val = "(%s) " % self.fixed[k].shortName()
                else :
                    val = "(%.1f) "%self.fixed[k]
                basename = re.sub( par, val, basename )
            i = 0
            par = "q_0 "
            for k in self.parlist :
                old = "p_%d "%k
                basename = re.sub( old, par, basename )
                i += 1
                par = "q_%d "%i
            basename = re.sub( "q_", "p_", basename )

        if npars == 0 : return basename
        k = 0
        par0 = "p_0"
        par1 = "q_%d"%(k+npars)
        while re.search( par0, basename ) :
            basename = re.sub( par0, par1, basename )
            k += 1
            par0 = "p_%d "%k
            par1 = "q_%d "%(k+npars)
        basename = re.sub( "q_", "p_", basename )
        return basename





#!/usr/bin/env python

import numpy as numpy
import math as math
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from numpy.testing import assert_array_almost_equal as assertAAE

# This file is part of the BayesicFitting package.
#
# BayesicFitting is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or ( at your option ) any later version.
#
# BayesicFitting is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#
# 2016 Do Kester

def plotFit( x, data=None, model=None, ftr=None, truth=None, show=True,
             residuals=False ) :

    minx = numpy.min( x )
    maxx = numpy.max( x )
    np = 1
    plt.figure( "fit" )
    ax0 = plt

    if residuals :
        plt.subplots_adjust( hspace=0.001 )
        np = 2
        gs = gridspec.GridSpec( 2, 1, height_ratios=[4, 1])

        ax1 = plt.subplot( gs[1] )
        yfit = model( x )
        res = data - yfit
        ax1.plot( x, res, 'k-' )
        ax1.margins( 0.05, 0.05 )
        plt.ylabel( "residual" )
        plt.xlabel( "x" )

        ax0 = plt.subplot( gs[0] )
        xticklabels = ax0.get_xticklabels()
        plt.setp( xticklabels, visible=False )

    if truth is not None :
        ax0.plot( x, truth, 'b+' )
    if data is not None :
        ax0.plot( x, data, 'k.' )

    if model is not None :
        xx = numpy.linspace( minx, maxx, 10 * len( x ) )
        yy = model( xx )
        if ftr is not None :
            err = ftr.monteCarloError( xx )
            ax0.plot( xx, yy - err, 'g-' )
            ax0.plot( xx, yy + err, 'g-' )
        ax0.plot( xx, yy, 'r-' )

    ax0.margins( 0.05, 0.05 )
    plt.ylabel( "y" )
    if not residuals :
        plt.xlabel( "x" )

    if show :
        plt.show()

def plotModel( model, par ) :
    xx = numpy.linspace( -1, +1, 1001 )
    plt.plot( xx, model.result( xx, par ), '-', linewidth=2 )
    x2 = numpy.linspace( -1, +1, 11 )
    dy = model.derivative( x2, par )
    yy = model.result( x2, par )
    for k in range( 11 ) :
        x3 = numpy.asarray( [-0.05, +0.05] )
        y3 = x3 * dy[k] + yy[k]
        plt.plot( x2[k] + x3, y3, 'r-' )

    plt.show()




def plotErrdis( errdis, model, limits=[-10,10], max=None, plot=None ) :
    np = 2001
    map = numpy.ndarray( np, dtype=float )

    par = model.parameters
    pars = numpy.append( par, errdis.hypar )
    r0 = numpy.linspace( limits[0], limits[1], np, dtype=float )

    for k, p in enumerate( r0 ) :
        pars[0] = p
        map[k] = errdis.logLikelihood( model, pars )

    if max is None :
        max = numpy.max( map )

    map = numpy.exp( map - max )
    if plot :
        plt.figure( str( errdis ) )
        plt.plot( r0, map, 'k-' )

    return ( math.log( numpy.sum( map ) / np ) + max, max )


def plotErrdis2d( errdis, model, limits=[-10,10], nslim=None, max=None,
                plot=None ) :

    np = 201
    map = numpy.ndarray( (np,np), dtype=float )

    par = model.parameters
    pars = numpy.append( par, errdis.hypar )
    r0lo = limits[0]
    r0hi = limits[1]
    r0 = numpy.linspace( r0lo, r0hi, np, dtype=float )
    if nslim is None :
        r1lo, r1hi = ( r0lo, r0hi )
        r1 = numpy.linspace( r1lo, r1hi, np, dtype=float )
        prior1 = numpy.zeros_like( r1 ) + 1 / ( r1hi - r1lo )
    else :
        r1lo, r1hi = ( nslim[0], nslim[1] )
        r1 = numpy.linspace( r1lo, r1hi, np, dtype=float )
        cp = math.log( r1hi / r1lo )                # integral of Jeffreys
        prior1 = 1 / ( cp * r1 )


    pixsz = ( r0hi - r0lo ) * ( r1hi - r1lo ) / ( np * np )
    prior0 = numpy.zeros_like( r0 ) + 1 / ( r0hi - r0lo )
    prior = numpy.outer( prior1, prior0 ) * pixsz
    zp = numpy.sum( prior )
#    print( numpy.sum( prior0 ), numpy.sum( prior1 ), zp )
    assertAAE( zp, 1.0, 2 )

#    print( prior.shape, prior[0,0], ( 1.0/ ( r0hi - r0lo ) ) )

    for k1,p1 in enumerate( r1 ) :
        pars[1] = p1
        for k0,p0 in enumerate( r0 ) :
            pars[0] = p0
            map[k1,k0] = errdis.logLikelihood( model, pars )

    logl = numpy.max( map )
#    print( prior.shape, prior[0,0], 2*math.log( 1.0/ ( r0hi - r0lo ) ) )
#    print( pixsz )

    map += numpy.log( prior )

    if max is None :
        max = numpy.max( map )

    map = numpy.exp( map - max )

    if plot :
        plt.figure( str( errdis ) )

        # More colormaps in matplotlib: docs->Matplotlib examples->color Examples

        plt.imshow( map, aspect=None, cmap='viridis', origin='lower',
                extent=(r0lo,r0hi,r1lo,r1hi) )
        plt.xlabel( "parameter 0" )
        plt.ylabel( "parameter 1" )

    return ( ( math.log( numpy.sum( map ) ) + max ), logl )






#!/usr/bin/env python
# -*- coding: utf-8 -*-
# PYTHON.test -> test area for basic fitter package

"""
from Test2dFixedModels import Test2dFixedModels
from Test2dModels import Test2dModels
# from TestAll import TestAll
from TestAmoeba import TestAmoeba
from TestAmoebaFitter import TestAmoebaFitter
from TestBaseFitter import TestBaseFitter
from TestBaseModel import TestBaseModel
from TestBracketModel import TestBracketModel
from TestCombiModel import TestCombiModel
from TestCompoundModel import TestCompoundModel
from TestCurveFitter import TestCurveFitter
from TestDriftModels import TestDriftModels
from TestEngine import TestEngine
from TestEngine2 import TestEngine2
from TestEngine3 import TestEngine3
from TestErrorDistribution import TestErrorDistribution
from TestEvidence import TestEvidence
from TestEvidence2 import TestEvidence2
from TestFitter import TestFitter
from TestFixedModels import TestFixedModels
from TestFormatter import TestFormatter
from TestFreeShape2dModel import TestFreeShape2dModel
from TestFreeShapeModel import TestFreeShapeModel
from TestImageAssistant import TestImageAssistant
from TestKernel2dModel import TestKernel2dModel
from TestKernels import TestKernels
from TestLMFitter import TestLMFitter
from TestLogFactorial import TestLogFactorial
from TestModels import TestModels
from TestMonteCarlo import TestMonteCarlo
from TestNLFitters import TestNLFitters
from TestNestedSampler import TestNestedSampler
from TestPrior import TestPrior
from TestQRFitter import TestQRFitter
from TestRobustShell import TestRobustShell
from TestSampleList import TestSampleList
from TestStartEngine import TestStartEngine
from TestTools import TestTools
from TestWeights import TestWeights

print( "__init__  done" )
"""
